@@ -3,7 +3,7 @@ package Term::ReadLine::Perl5;
 use warnings; use strict;
 no warnings 'once';
 
-our $VERSION = '1.13';
+our $VERSION = '1.20';
 
 =head1 NAME
 
@@ -34,9 +34,12 @@ Eval, Print Loops).
 =cut
 
 use Carp;
+use Term::ReadLine::Perl5::History;
+use Term::ReadLine::Perl5::Tie;
+use Term::ReadLine::Perl5::readline;
 
-our @ISA = qw(Term::ReadLine::Stub Term::ReadLine::Perl5::AU);
-my (%attribs, @history, $term);
+our @ISA = qw(Term::ReadLine::Stub);
+my (%attribs, $term);
 
 my %features = (
 		 appname => 1,       # "new" is recognized
@@ -63,29 +66,9 @@ my %features = (
 
 sub readline {
   shift;
-  &readline::readline(@_);
+  &Term::ReadLine::Perl5::readline::readline(@_);
 }
 
-# add_history is what GNU ReadLine defines. AddHistory is what we have
-# below.
-*add_history = \&AddHistory;
-
-# Not sure if addhistory() is needed. It is possible it was misspelling
-# of add_history.
-*addhistory = \&AddHistory;
-
-# for backward compatibility: StifleHistory is the old name.
-*StifleHistory = \&stifle_history;
-
-# Initializations of variables to pacify -w
-$readline::minlength = 1;
-$readline::rl_readline_name = undef;
-$readline::rl_basic_word_break_characters = undef;
-$readline::history_stifled = 0;
-$readline::rl_history_length = 0;
-$readline::rl_max_input_history = 0;
-
-
 =head2
 
 C<Term::ReadLine::Perl->new($name, [*IN, [*OUT])>
@@ -123,7 +106,7 @@ sub new {
       warn "Ignoring name of second readline interface.\n" if defined $term;
       shift;
     } else {
-      $readline::rl_readline_name = shift; # Name
+      $Term::ReadLine::Perl5::readline::rl_readline_name = shift; # Name
     }
   }
   if (!@_) {
@@ -136,17 +119,17 @@ sub new {
 	(($IN eq 'CONIN$' and $^O eq 'MSWin32') ? "+< $IN" : "< $IN")
 	  or croak "Cannot open $IN for read";
       open(OUT,">$OUT") || croak "Cannot open $OUT for write";
-      $readline::term_IN = \*IN;
-      $readline::term_OUT = \*OUT;
+      $Term::ReadLine::Perl5::readline::term_IN = \*IN;
+      $Term::ReadLine::Perl5::readline::term_OUT = \*OUT;
     }
   } else {
     if (defined $term and ($term->IN ne $_[0] or $term->OUT ne $_[1]) ) {
       croak "Request for a second readline interface with different terminal";
     }
-    $readline::term_IN = shift;
-    $readline::term_OUT = shift
+    $Term::ReadLine::Perl5::readline::term_IN = shift;
+    $Term::ReadLine::readline::term_OUT = shift
   }
-  eval {require Term::ReadLine::readline}; die $@ if $@;
+  eval {require Term::ReadLine::Perl5::readline}; die $@ if $@;
   # The following is here since it is mostly used for perl input:
   # $readline::rl_basic_word_break_characters .= '-:+/*,[])}';
   $term = bless [$readline::term_IN,$readline::term_OUT];
@@ -155,14 +138,14 @@ sub new {
     local $SIG{__WARN__} = sub {}; # With older Perls
     $term->ornaments(1);
   }
-  $readline::rl_history_length = $readline::rl_max_input_history = 0;
+  $rl_history_length = $rl_max_input_history = 0;
   return $term;
 }
 
 sub newTTY {
   my ($self, $in, $out) = @_;
-  $readline::term_IN   = $self->[0] = $in;
-  $readline::term_OUT  = $self->[1] = $out;
+  $Term::ReadLine::Perl5::readline::term_IN   = $self->[0] = $in;
+  $Term::ReadLine::Perl5::readline::term_OUT  = $self->[1] = $out;
   my $sel = select($out);
   $| = 1;				# for DB::OUT
   select($sel);
@@ -170,6 +153,32 @@ sub newTTY {
 
 sub ReadLine {'Term::ReadLine::Perl5'}
 
+=head2 stifle_history
+
+C<stifle_history($max)>
+
+Stifle or put a cap on thethe history list, remembering only C<$max>
+number of lines.
+
+=cut
+
+### FIXME: stifle_history is still here because it updates $attribs.
+## Pass a reference?
+sub stifle_history {
+  shift;
+  my $max = shift;
+  $max = 0 if !defined($max) || $max < 0;
+
+  if (scalar @rl_History > $max) {
+      splice @rl_History, $max;
+      $attribs{history_length} = scalar @rl_History;
+  }
+
+  $history_stifled = 1;
+  $rl_max_input_history = $max;
+}
+
+
 =head2
 
 C<MinLine([$minlength])>
@@ -182,65 +191,40 @@ The previous value is returned.
 =cut
 
 sub MinLine {
-    my $old = $readline::minlength;
-    $readline::minlength = $_[1] if @_ == 2;
+    my $old = $minlength;
+    $minlength = $_[1] if @_ == 2;
     return $old;
 }
 
-sub SetHistory {
-    shift;
-    @readline::rl_History = @_;
-    $readline::rl_HistoryIndex = @readline::rl_History;
-    $readline::rl_history_length = $readline::rl_max_input_history =
-	@readline::rl_History;
+sub Features { \%features; }
+
+tie %attribs, 'Term::ReadLine::Perl5::Tie' or die ;
+sub Attribs {
+  \%attribs;
 }
+#################### History ##########################################
 
-sub GetHistory {
-    @readline::rl_History;
-}
+# GNU ReadLine names
+*add_history            = \&Term::ReadLine::Perl5::History::add_history;
+*clear_history          = \&Term::ReadLine::Perl5::History::clear_history;
+*history_list           = \&Term::ReadLine::Perl5::History::history_list;
+*history_is_stifled     = \&Term::ReadLine::Perl5::History::history_is_stifled;
+*read_history           = \&Term::ReadLine::Perl5::History::read_history;
+*replace_history_entry  = \&Term::ReadLine::Perl5::History::replace_history_entry;
+*unstifle_history       = \&Term::ReadLine::Perl5::History::unstifle_history;
+*write_history          = \&Term::ReadLine::Perl5::History::write_history;
 
-=head2 AddHistory
+# Some Term::ReadLine::Gnu names
+*AddHistory             = \&Term::ReadLine::Perl5::History::AddHistory;
+*GetHistory             = \&Term::ReadLine::Perl5::History::GetHistory;
+*ReadHistory            = \&Term::ReadLine::Perl5::History::ReadHistory;
+*SetHistory             = \&Term::ReadLine::Perl5::History::SetHistory;
+*WriteHistory           = \&Term::ReadLine::Perl5::History::WriteHistory;
 
-C<AddHistory($line1, ...)>
+# Backward compatibility:
+*addhistory = \&Term::ReadLine::Perl5::add_History;
+*StifleHistory = \&stifle_history;
 
-Place @_ at the end of the history list unless the history is stifled,
-or there are already too many items.
-
-=cut
-
-sub AddHistory {
-    shift;
-    if ($readline::history_stifled &&
-	($readline::rl_history_length == $readline::rl_max_input_history)) {
-	# If the history is stifled, and history_length is zero,
-	# and it equals max_input_history, we don't save items.
-	return if $readline::rl_max_input_history == 0;
-	shift @readline::rl_History;
-    }
-
-    push @readline::rl_History, @_;
-    $readline::rl_HistoryIndex = @readline::rl_History + @_;
-    $readline::rl_history_length = scalar @readline::rl_History;
-}
-
-=head2 clear_history
-
-C<clear_history()>
-
-Clear or reset readline history.
-
-=cut
-
-sub clear_history {
-  shift;
-  @readline::rl_History = ();
-  $readline::rl_HistoryIndex = $readline::rl_history_length = 0;
-}
-
-sub history_list
-{
-    @readline::rl_History[1..$#readline::rl_History]
-}
 
 =head2 remove_history
 
@@ -255,160 +239,17 @@ sub remove_history {
   shift;
   my $which = $_[0];
   return undef if
-    $which < 0 || $which >= $readline::rl_history_length ||
+    $which < 0 || $which >= $rl_history_length ||
       $attribs{history_length} ==  0;
-  my $removed = splice @readline::rl_History, $which, 1;
-  $readline::rl_history_length--;
-  $readline::rl_HistoryIndex = $readline::rl_history_length if
-    $readline::rl_history_length < $readline::rl_HistoryIndex;
+  my $removed = splice @rl_History, $which, 1;
+  $rl_history_length--;
+  $rl_HistoryIndex =
+      $rl_history_length if
+    $rl_history_length <
+    $rl_HistoryIndex;
   return $removed;
 }
 
-=head2 replace_history_entry
-
-C<replace_history_entry($which, $data)>
-
-Make the history entry at C<$which> have C<$data>.  This returns the old
-entry. In the case of an invalid C<$which>, $<undef> is returned.
-
-=cut
-
-sub replace_history_entry {
-  shift;
-  my ($which, $data) = @_;
-  return undef if $which < 0 || $which >= $readline::rl_history_length;
-  my $replaced = splice @readline::rl_History, $which, 1, $data;
-  return $replaced;
-}
-
-=head2 stifle_history
-
-C<stifle_history($max)>
-
-Stifle or put a cap on thethe history list, remembering only C<$max>
-number of lines.
-
-=cut
-
-sub stifle_history {
-  shift;
-  my $max = shift;
-  $max = 0 if !defined($max) || $max < 0;
-
-  if (scalar @readline::rl_History > $max) {
-      splice @readline::rl_History, $max;
-      $attribs{history_length} = scalar @readline::rl_History;
-  }
-
-  $readline::history_stifled = 1;
-  $readline::rl_max_input_history = $max;
-}
-
-=head2 unstifle_history
-
-C<unstifle_history>
-
-Unstifle or remove limit the history list.
-
-Theprevious maximum number of history entries is returned.  The value
-is positive if the history was stifled and negative if it wasn't.
-
-=cut
-
-sub unstifle_history {
-  if ($readline::history_stifled) {
-    $readline::history_stifled = 0;
-    return (scalar @readline::rl_History);
-  } else {
-    return - scalar @readline::rl_History;
-  }
-}
-
-=head2 history_is_stifled
-
-C<history_is_stifled>
-
-Returns I<true> if saved history has a limited (stifled) or I<false>
-if there is no limit (unstifled).
-
-=cut
-
-sub history_is_stifled {
-  shift;
-  $readline::history_stifled ? 1 : 0;
-}
-
-# read_history() and write_history() follow GNU Readline's
-# C convention of returning 0 for success and 1 for failure.
-#
-# ReadHistory and WriteHstory follow Perl's convention of returning 1
-# for success and 0 for failure.
-# It is a little bit whacky, but this is in fact how Term::ReadLine::Gnu
-# works.
-
-sub read_history {
-  my $self = shift;
-  my $filename = shift;
-  open(HISTORY, '<', $filename ) or return 1;
-  while (<HISTORY>) { chomp; push @history, $_} ;
-  SetHistory($self, @history);
-  close HISTORY;
-  return 0;
-}
-
-sub write_history {
-  shift;
-  my $filename = shift;
-  open(HISTORY, '>', $filename ) or return 1;
-  for (@readline::rl_History) { print HISTORY $_, "\n"; }
-  close HISTORY;
-  return 0;
-}
-
-sub ReadHistory {
-    ! read_history(@_);
-}
-
-sub WriteHistory {
-    ! write_history(@_);
-}
-
-sub Features { \%features; }
-
-tie %attribs, 'Term::ReadLine::Perl5::Tie' or die ;
-sub Attribs {
-  \%attribs;
-}
-sub DESTROY {}
-
-package Term::ReadLine::Perl5::AU;
-
-sub AUTOLOAD {
-  my $AUTOLOAD =~ s/.*:://; 		# preserve match data
-  my $name = "readline::rl_$AUTOLOAD";
-  die "Unknown method `$AUTOLOAD' in Term::ReadLine::Perl5"
-    unless exists $readline::{"rl_$AUTOLOAD"};
-  *$AUTOLOAD = sub { shift; &$name };
-  goto &$AUTOLOAD;
-}
-
-package Term::ReadLine::Perl5::Tie;
-
-sub TIEHASH { bless {} }
-sub DESTROY {}
-
-sub STORE {
-  my ($self, $name) = (shift, shift);
-  no strict;
-  $ {'readline::rl_' . $name} = shift;
-}
-
-sub FETCH {
-  my ($self, $name) = (shift, shift);
-  no strict;
-  $ {'readline::rl_' . $name};
-}
-
 1;
 
 __END__
